#define _GNU_SOURCE
#define _FILE_OFFSET_BITS 64
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include "md5.h"

long long offset;
long long length;
char md5[33];

void usage (void)
{
	fprintf(stderr,
	  "\n"
	  "Usage:\n"
	  "\n"
	  "cdtest -c file\n"
	  "    Calculate and write size and md5sum metainfo to the file.\n"
	  "    file must be seekable.\n"
	  "\n"
	  "cdtest -c infile outfile\n"
	  "    Copy infile to outfile and write size and md5sum metainfo\n"
	  "    into outfile.\n"
	  "    infile can be /dev/stdin or whatever you want.\n"
	  "    outfile must be seekable.\n"
	  "\n"
	  "cdtest [-p pid] [-s sig] [-f fd] [-n num] file\n"
	  "    Check the integrity of the file.\n"
	  "    -p pid:\n"
	  "        Send a signal to this process after every percent completed.\n"
	  "        This is disabled by default.\n"
	  "    -s sig:\n"
	  "         Send this kind of signal. Only numbers are accepted, names aren't.\n"
	  "         Default is 10 (SIGUSR1).\n"
	  "    -f fd:\n"
	  "         Write a # after each percent and a newline at the end to this fd.\n"
	  "         This is disabled by default.\n"
	  "    -n num:\n"
	  "         Instead of using percents (100), divide by this number.\n"
	  "    file can be /dev/stdin or /dev/cdrom or whatever you want.\n"
	  "\n"
	  );
}

/*
 *  Wrapper around read() and write().
 */
int readx (int fd, void *buf, size_t count)
{
	int i;
	int r = 0;
	while (count > 0) {
		i = read(fd, buf, count);
		if (i < 0) {
			if (errno == EINTR) continue;
			return -1;
		}
		if (i == 0) break;
		r += i;
		buf += i;
		count -= i;
	}
	return r;
}

int writex (int fd, const void *buf, size_t count)
{
	int i;
	int r = 0;
	while (count > 0) {
		i = write(fd, buf, count);
		if (i < 0) {
			if (errno == EINTR) continue;
			return -1;
		}
		if (i == 0) break;
		r += i;
		buf += i;
		count -= i;
	}
	return r;
}

/*
 *  s points to the first 65536 bytes of the image.
 *  Find the offset where the file length and md5 is stored. This is based on
 *  the source of isodebug (mkisofs).
 *  Read the file length and md5.
 *  Return -1 in case of error, the offset in case of success.
 */
int get_metainfo (const unsigned char *s)
{
	int i, j;
	unsigned char zero[32];
	for (i = 32768; i < 63488; i += 2048) {
		if (s[i] == 255) break;
	}
	if (i == 63488) return -1;
	i += 2048;
	if (strncmp(&s[i], "MKI ", 4)) return -1;
	j = strnlen(&s[i], 2000);
	if (j == 2000) return -1;
	i += j+1;
	for (j = 0; j < 32; j++) zero[j] = 0;
	if (s[i+12] != 0) return -1;
	if (s[i+12+1+32] != 0) return -1;
	if (!memcmp(&s[i], zero, 12) && !memcmp(&s[i+12+1], zero, 32)) {
		length = -1;
		md5[0] = 0;
		return i;
	}
	for (j = 0; j < 12; j++) {
		if (s[i+j] < '0' || s[i+j] > '9') return -1;
	}
	for (j = 0; j < 32; j++) {
		if ((s[i+12+1+j] < '0' || s[i+12+1+j] > '9')
		 && (s[i+12+1+j] < 'a' || s[i+12+1+j] > 'f')) return -1;
	}
	sscanf(&s[i], "%lld", &length);
	strcpy(md5, &s[i+12+1]);
	return i;
}

/*
 *  Everything else.
 */
int main (int argc, char *argv[])
{
	int fd, fd2;
	long long offset;
	long long off2;
	long long toread;
	long long steps, steps_total;
	long long trailingzeros;
	int j, j2;

	unsigned char buf[307200];

	int calc = 0;
	int copy = 0;
	pid_t pid = -1;
	int sig = SIGUSR1;
	int f = -1;
	int n = 100;

	struct md5_ctx ctx;
	unsigned char md5raw[16];
	char md5ascii[33];

	int i, i2;
	int c;
	while ((c = getopt(argc, argv, "cf:n:p:s:")) != EOF) {
		switch (c) {
		  case 'c':
			calc = 1;
			break;
		  case 'f':
			f = atoi(optarg);
			break;
		  case 'n':
			n = atoi(optarg);
			break;
		  case 'p':
			pid = atoi(optarg);
			break;
		  case 's':
			sig = atoi(optarg);
			break;
		  case '?':
			usage();
			exit(1);
		}
	}
	if (calc && argc == optind + 2) {
		copy = 1;
	} else if (argc != optind + 1) {
		usage();
		exit(1);
	}

	fd = open(argv[optind], (calc && !copy ? O_RDWR : O_RDONLY) | O_LARGEFILE);
	if (fd < 0) {
		fprintf(stderr, "Cannot open %s: %s\n", argv[optind], strerror(errno));
		exit(1);
	}
	if (copy) {
		fd2 = open(argv[optind + 1], O_WRONLY | O_CREAT | O_TRUNC | O_LARGEFILE, 00666);
		if (fd2 < 0) {
			fprintf(stderr, "Cannot open %s: %s\n", argv[optind + 1], strerror(errno));
			exit(1);
		}
	} else {
		fd2 = fd;
	}
	i = readx(fd, buf, 65536);
	if (i < 0) {
		fprintf(stderr, "Read error: %s\n", strerror(errno));
		exit(1);
	}
	if (copy) {
		i2 = writex(fd2, buf, i);
		if (i2 != i) {
			fprintf(stderr, "Write error: %s\n", i2 < 0 ? strerror(errno) : "short write");
			exit(1);
		}
	}
	if (i != 65536 || (offset = get_metainfo(buf)) < 0) {
		fprintf(stderr, "This doesn't seem to be an ISO image generated by mkisofs.\n");
		exit(1);
	}

	if (calc) {
		if (length >= 0) {
			fprintf(stderr, "This image already contains md5 information.\n");
			exit(1);
		}
		__md5_init_ctx(&ctx);
		__md5_process_bytes(buf, 65536, &ctx);
		off2 = 65536;
		trailingzeros = 0;
		while (1) {
			i = readx(fd, buf, 65536);
			if (i < 0) {
				fprintf(stderr, "Read error: %s\n", strerror(errno));
				exit(1);
			}
			if (i == 0) break;
			if (copy) {
				i2 = writex(fd2, buf, i);
				if (i2 != i) {
					fprintf(stderr, "Write error: %s\n", i2 < 0 ? strerror(errno) : "short write");
					exit(1);
				}
			}
			__md5_process_bytes(buf, i, &ctx);
			off2 += i;
			j = i - 1;
			while (j >= 0 && buf[j] == 0) j--;
			if (j < 0) trailingzeros += i;
			else trailingzeros = i - 1 - j;
			if (i != 65536) break;
		}
		__md5_finish_ctx(&ctx, md5raw);
		if (copy) {
			close(fd);
		}
		if (off2 < 65536 + 307200) {
			fprintf(stderr, "Image too short.\n");
			exit(1);
		}
		if (trailingzeros < 307200) {
			fprintf(stderr, "Image doesn't end with 307200 zero bytes.\n");
			exit(1);
		}
		for (i = 0; i < 16; i++) {
			sprintf(&md5ascii[2*i], "%02x", md5raw[i]);
		}
		sprintf(buf, "%012lld", off2);
		i = lseek(fd2, offset, SEEK_SET);
		if (i == -1) {
			fprintf(stderr, "Seek error: %s\n", strerror(errno));
			exit(1);
		}
		i = writex(fd2, buf, 12);
		if (i != 12) {
			fprintf(stderr, "Write error: %s\n", i < 0 ? strerror(errno) : "Short write");
			exit(1);
		}
		i = lseek(fd2, offset+12+1, SEEK_SET);
		if (i == -1) {
			fprintf(stderr, "Seek error: %s\n", strerror(errno));
			exit(1);
		}
		i = writex(fd2, md5ascii, 32);
		if (i != 32) {
			fprintf(stderr, "Write error: %s\n", i < 0 ? strerror(errno) : "Short write");
			exit(1);
		}
		close(fd2);
		printf("Metainfo successfully written.\n");
		return 0;
	}

	if (length < 0) {
		fprintf(stderr, "This image doesn't contain md5 information.\n");
		exit(1);
	}
	if (length < 65536 + 307200) {
		fprintf(stderr, "Image too short.\n");
		exit(1);
	}
	length -= 307200;
	__md5_init_ctx(&ctx);
	memset(&buf[offset], 0, 12);
	memset(&buf[offset+12+1], 0, 32);
	__md5_process_bytes(buf, 65536, &ctx);
	steps_total = length / 65536;
	steps = 1;
	j = 0;
	off2 = 65536;
	while (off2 < length) {
		toread = length - off2;
		if (toread >= 65536) toread = 65536;
		i = readx(fd, buf, toread);
		if (i < 0) {
			fprintf(stderr, "Read error: %s\n", strerror(errno));
			exit(1);
		}
		if (i != toread) {
			if (f >= 0) write(f, "\n", 1);
			fprintf(stderr, "Error: Premature end of file.\n");
			exit(1);
		}
		__md5_process_bytes(buf, i, &ctx);
		off2 += i;
		steps++;
		j2 = (long long)n * steps / steps_total;
		for (; j < j2; j++) {
			if (pid > 0) kill(pid, sig);
			if (f >= 0) write(f, "#", 1);
		}
	}
	memset(buf, 0, 307200);
	__md5_process_bytes(buf, 307200, &ctx);
	__md5_finish_ctx(&ctx, md5raw);
	close(fd);
	for (; j < n; j++) {
		if (pid > 0) kill(pid, sig);
		if (f >= 0) write(f, "#", 1);
	}
	if (f >= 0) write(f, "\n", 1);
	for (i = 0; i < 16; i++) {
		sprintf(&md5ascii[2*i], "%02x", md5raw[i]);
	}
	if (memcmp(md5ascii, md5, 32)) {
		fprintf(stderr, "Error: md5sum differs.\n");
		exit(1);
	}
	printf("ISO file is okay.\n");
	return 0;
}
